#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "cdbmake.h"
#include "logmsg.h"
#include "open.h"
#include "buffer.h"
#include "stralloc.h"
#include "getln.h"
#include "exit.h"
#include "byte.h"

#define WHO "ofmipname"

int rename(const char *,const char *); // stdio.h

char *fncdb;
char *fntmp;

void nomem()
{ logmsg(WHO,111,FATAL,"out of memory"); }
void die_read()
{ logmsg(WHO,111,FATAL,"unable to read input: "); }
void die_write()
{ logmsg(WHO,111,FATAL,B("unable to create: ",fntmp)); }
void usage()
{ logmsg(WHO,100,USAGE,"ofmipname cdb tmp"); }

struct cdb_make cdb;

stralloc line = {0};
int match;

stralloc key = {0};
stralloc data = {0};

void doit(char *x,unsigned int len)
{
  unsigned int colon;

  if (!len) return;
  if (x[0] == '#') return;

  colon = byte_chr(x,len,':');
  if (colon == len) return;
  if (!stralloc_copyb(&key,x,colon)) nomem();
  ++colon; x += colon; len -= colon;

  colon = byte_chr(x,len,':');
  if (colon == len) return;
  if (!stralloc_copyb(&data,x,colon)) nomem();
  if (!stralloc_0(&data)) nomem();
  ++colon; x += colon; len -= colon;

  colon = byte_chr(x,len,':');
  if (colon == len) return;
  if (!stralloc_catb(&data,x,colon)) nomem();
  ++colon; x += colon; len -= colon;

  if (cdb_make_add(&cdb,key.s,key.len,data.s,data.len) == -1)
    die_write();
}

int main(int argc,char **argv)
{
  int fd;

  umask(033);

  fncdb = argv[1];
  if (!fncdb) usage();
  fntmp = argv[2];
  if (!fntmp) usage();

  fd = open_trunc(fntmp);
  if (fd == -1) die_write();
  if (cdb_make_start(&cdb,fd) == -1) die_write();

  for (;;) {
    if (getln(buffer_0,&line,&match,'\n') == -1) die_read();
    doit(line.s,line.len);
    if (!match) break;
  }

  if (cdb_make_finish(&cdb) == -1) die_write();
  if (fsync(fd) == -1) die_write();
  if (close(fd) == -1) die_write(); /* NFS stupidity */
  if (rename(fntmp,fncdb) == -1)
    logmsg(WHO,111,FATAL,B("unable to move ",fntmp," to ",fncdb));

  _exit(0);
}
